from __future__ import annotations

import re
import os
import shutil

from pathlib import Path
from datetime import datetime, timezone
from typing import cast, Any, TypeVar, Generic, TypedDict, Callable
from threading import Lock
from yaml import safe_load, safe_dump
from xml.etree.ElementTree import Element

from ...xml import encode
from ..reporter import Reporter


CURRENT_STATE_VERSION = "1.0.1"
S = TypeVar("S")

_STATE_FILE = "state.yaml"

class _StateRoot(TypedDict):
  version: str
  created_at: str
  updated_at: str
  payload: Any


# thread safe
class Context(Generic[S]):
  def __init__(self, reporter: Reporter, path: Path, init: Callable[[], S]) -> None:
    self._reporter = reporter
    self._state: S
    self._state_lock: Lock = Lock()
    self._path: Path = path
    self._created_at: str

    state: S | None = None
    created_at: str | None = None
    if path.exists():
      assert path.is_dir(), f"Path {path} is not a directory"
      state_path = path.joinpath(_STATE_FILE)
      if state_path.exists():
        state, created_at = self._load_state(state_path)
      if state is None:
        shutil.rmtree(path)

    if state is None:
      path.mkdir(parents=True)
      state = init()
      created_at = self._current_utc()

    self._state = state
    self._created_at = created_at

  def _load_state(self, state_path: Path) -> tuple[S, str] | tuple[None, None]:
    with state_path.open("r", encoding="utf-8") as file:
      root = cast(_StateRoot, safe_load(file))
      version = root["version"]
      if version != CURRENT_STATE_VERSION:
        return None, None
      return cast(S, root["payload"]), root["created_at"]


  @property
  def reporter(self) -> Reporter:
    return self._reporter

  @property
  def path(self) -> Path:
    return self._path

  @property
  def state(self) -> S:
    with self._state_lock:
      return self._state

  @state.setter
  def state(self, state: S) -> None:
    with self._state_lock:
      self._state = state
      self.atomic_write(
        file_path=self._path / _STATE_FILE,
        content=safe_dump({
          "version": CURRENT_STATE_VERSION,
          "created_at": self._created_at,
          "updated_at": self._current_utc(),
          "payload": self._state,
        }),
      )

  def write_xml_file(self, file_path: Path, xml: Element) -> None:
    file_content = encode(xml)
    base_path = file_path.parent
    if not base_path.exists():
      base_path.mkdir(parents=True)
    self.atomic_write(file_path, file_content)

  def _current_utc(self) -> str:
    return datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S")

  def _split_index_suffix(self, file_name: str) -> tuple[str | None, int, int]:
    matches = re.match(r"^[a-zA-Z]+_\d+(_\d+)?\.xml$", file_name)
    if not matches:
      return None, -1, -1

    file_prefix: str
    index1: str
    index2: str
    cells = re.sub(r"\..*$", "", file_name).split("_")
    if len(cells) == 3:
      file_prefix, index1, index2 = cells
    else:
      file_prefix, index1 = cells
      index2 = index1

    return file_prefix, int(index1), int(index2)

  def atomic_write(self, file_path: Path, content: str):
    try:
      with open(file_path, "w", encoding="utf-8") as file:
        file.write(content)
        file.flush()
    except Exception as e:
      if os.path.exists(file_path):
        os.unlink(file_path)
      raise e



from pathlib import Path
from xml.etree.ElementTree import Element
from shutil import rmtree

from ...llm import LLM
from ..reporter import Reporter, AnalysingStep
from ..contents import Contents, Chapter
from ..data import ASSET_LAYOUT_KINDS, Paragraph
from ..utils import xml_files, Context, MultiThreads
from .common import State, Phase
from .contents_mapper import map_contents
from .patcher import read_paragraphs, read_paragraphs_with_patches


def generate_chapters(
      llm: LLM,
      reporter: Reporter,
      threads: MultiThreads,
      contents: Contents | None,
      sequence_path: Path,
      workspace_path: Path,
      max_request_tokens: int,
    ) -> tuple[Path, Contents | None]:

  map_path: Path = workspace_path / "map"
  output_path = workspace_path / "output"
  context: Context[State] = Context(
    reporter=reporter,
    path=workspace_path,
    init=lambda: {
      "phase": Phase.MAPPER.value,
      "has_contents": False,
      "max_request_tokens": max_request_tokens,
      "completed_ranges": [],
    },
  )
  if context.state["phase"] == Phase.MAPPER:
    has_contents = False
    if contents is not None:
      has_contents = True
      context.reporter.go_to_step(AnalysingStep.MAPPING_CONTENTS)
      map_contents(
        llm=llm,
        context=context,
        threads=threads,
        contents=contents,
        sequence_path=sequence_path,
        map_path=map_path,
      )

    context.state = {
      **context.state,
      "phase": Phase.CHAPTER.value,
      "has_contents": has_contents,
    }

  used_chapter_ids: set[int] = set()

  if context.state["phase"] == Phase.CHAPTER:
    contents_and_map_path: tuple[Contents, Path] | None = None
    if contents and context.state["has_contents"]:
      contents_and_map_path = (contents, map_path)

    rmtree(output_path, ignore_errors=True)
    _generate_chapter_xmls(
      context=context,
      contents_and_map_path=contents_and_map_path,
      used_chapter_ids=used_chapter_ids,
      sequence_path=sequence_path,
      output_path=output_path,
    )
    context.state = {
      **context.state,
      "phase": Phase.COMPLETED.value,
    }
  elif output_path.exists():
    for _, _, chapter_id, _ in xml_files(output_path):
      used_chapter_ids.add(chapter_id)

  if contents is not None:
    contents = _filter_contents(
      contents=contents,
      used_chapter_ids=used_chapter_ids,
    )
  return output_path, contents

def _generate_chapter_xmls(
      context: Context[State],
      contents_and_map_path: tuple[Contents, Path] | None,
      used_chapter_ids: set[int],
      sequence_path: Path,
      output_path: Path,
    ):

  chapter: Chapter | None = None
  chapter_element = Element("chapter")

  def save_chapter():
    nonlocal chapter, chapter_element
    if len(chapter_element) == 0:
      return

    file_name = "chapter.xml"
    if chapter is not None:
      file_name = f"chapter_{chapter.id}.xml"
      used_chapter_ids.add(chapter.id)

    context.write_xml_file(
      file_path=output_path / file_name,
      xml=chapter_element,
    )
    chapter_element = Element("chapter")

  if contents_and_map_path is None:
    for paragraph in read_paragraphs(sequence_path):
      for layout_element in _flat_layouts_from_paragraph(paragraph):
        chapter_element.append(layout_element)
  else:
    contents, map_path = contents_and_map_path

    for this_chapter, paragraph in read_paragraphs_with_patches(
      paragraph_path=sequence_path,
      contents=contents,
      map_path=map_path,
    ):
      if chapter is None:
        if this_chapter is not None:
          save_chapter()
          chapter = this_chapter
      elif this_chapter is not None and this_chapter.id != chapter.id:
        save_chapter()
        chapter = this_chapter

      for layout_element in _flat_layouts_from_paragraph(paragraph):
        chapter_element.append(layout_element)

  save_chapter()

def _flat_layouts_from_paragraph(paragraph: Paragraph):
  layout_element: Element | None = None
  for layout in paragraph.layouts:
    if layout_element is not None and \
       layout_element.tag != layout.kind:
      yield layout_element
      layout_element = None

    if layout.kind in ASSET_LAYOUT_KINDS:
      asset_element = layout.to_xml()
      if asset_element.get("hash", None):
        # TODO: 临时做法，为了排除 hash 为空的情况而做的防御性工作。
        #       之所以会出现此现象，是因为 https://github.com/oomol-lab/pdf-craft/issues/238 。
        yield asset_element
      continue

    if layout_element is None:
      layout_element = Element(layout.kind.value, {
        "id": layout.id,
      })
    for line in layout.lines:
      layout_element.append(line.to_xml())

  if layout_element is not None:
    yield layout_element

def _filter_contents(contents: Contents, used_chapter_ids: set[int]) -> Contents:
  return Contents(
    page_indexes=contents.page_indexes,
    chapters=_filter_chapters(
      raw_chapters=contents.chapters,
      used_chapter_ids=used_chapter_ids,
    ),
    prefaces=_filter_chapters(
      raw_chapters=contents.prefaces,
      used_chapter_ids=used_chapter_ids,
    ),
  )

def _filter_chapters(raw_chapters: list[Chapter], used_chapter_ids: set[int]) -> list[Chapter]:
  chapters: list[Chapter] = []
  for raw_chapter in raw_chapters:
    chapter = _filter_chapter(raw_chapter, used_chapter_ids)
    if chapter is not None:
      chapters.append(chapter)
  return chapters

def _filter_chapter(raw_chapter: Chapter, used_chapter_ids: set[int]) -> Chapter | None:
  filtered_chapter: Chapter | None = None
  children: list[Chapter] = []

  for sub_chapter in raw_chapter.children:
    sub_chapter = _filter_chapter(sub_chapter, used_chapter_ids)
    if sub_chapter is not None:
      children.append(sub_chapter)

  if len(children) > 0 or raw_chapter.id in used_chapter_ids:
    filtered_chapter = Chapter(
      id=raw_chapter.id,
      name=raw_chapter.name,
      children=children,
    )
  return filtered_chapter


from pathlib import Path
from typing import Generator
from xml.etree.ElementTree import Element

from ..utils import read_xml_file, xml_files, Context
from ..data import (
  Paragraph,
  ParagraphType,
  Caption,
  FormulaLayout,
  AssetLayout,
  Line,
  Layout,
  LayoutKind,
)


def read_paragraphs(dir_path: Path, name: str = "paragraph") -> Generator[Paragraph, None, None]:
  for file_path, _name, page_index, order_index in xml_files(dir_path):
    if name == _name:
      element = read_xml_file(file_path)
      yield decode_paragraph(element, page_index, order_index)

def decode_paragraph(element: Element, page_index: int, order_index: int):
  return Paragraph(
    type=ParagraphType(element.get("type")),
    page_index=page_index,
    order_index=order_index,
    layouts=[decode_layout(e) for e in element],
  )

def decode_layout(element: Element) -> Layout:
  id: str = element.get("id")
  kind = LayoutKind(element.tag)
  page_index, order_index = id.split("/", maxsplit=1)
  caption_lines: list[Line] = []
  body_lines: list[Line] = [
    _decode_line(line_element)
    for line_element in element
    if line_element.tag == "line"
  ]
  for child in element:
    if child.tag != "caption":
      continue
    for line_element in child:
      if line_element.tag != "line":
        continue
      caption_lines.append(_decode_line(line_element))

  if kind == LayoutKind.FORMULA:
    hash_hex = element.get("hash", "")
    latex_element = element.find("latex")
    latex_text: str = ""
    if latex_element is not None:
      latex_text = latex_element.text or ""

    return FormulaLayout(
      kind=kind,
      hash=bytes.fromhex(hash_hex),
      page_index=int(page_index),
      order_index=int(order_index),
      caption=Caption(lines=caption_lines),
      lines=body_lines,
      latex=latex_text,
    )
  elif kind in (LayoutKind.FIGURE, LayoutKind.TABLE):
    hash_hex = element.get("hash", "")
    return AssetLayout(
      kind=kind,
      hash=bytes.fromhex(hash_hex),
      page_index=int(page_index),
      order_index=int(order_index),
      caption=Caption(lines=caption_lines),
      lines=body_lines,
    )
  else:
    return Layout(
      kind=kind,
      page_index=int(page_index),
      order_index=int(order_index),
      caption=Caption(lines=caption_lines),
      lines=body_lines,
    )

def _decode_line(element: Element) -> Line:
  return Line(
    text=(element.text or "").strip(),
    confidence=element.get("confidence", "1.0"),
  )

class ParagraphWriter:
  def __init__(self, context: Context, dir_path: Path, name: str = "paragraph"):
    self._name: str = name
    self._context: Context = context
    self._dir_path: Path = dir_path

    if not self._dir_path.exists():
      self._dir_path.mkdir(parents=True)
    if not self._dir_path.is_dir():
      raise ValueError(f"Path {self._dir_path} is not a directory")

  def write(self, paragraph: Paragraph) -> None:
    file_name = f"{self._name}_{paragraph.page_index}_{paragraph.order_index}.xml"
    self._context.write_xml_file(
      file_path=self._dir_path / file_name,
      xml=paragraph.to_xml(),
    )


from dataclasses import dataclass
from typing import Generator
from xml.etree.ElementTree import Element
from ..data import Layout


@dataclass
class Line:
  id: int
  text: str
  splitted: bool

  def to_xml(self) -> Element:
    line_xml = Element("line")
    line_xml.text = self.text
    line_xml.set("id", str(self.id))
    return line_xml

@dataclass
class _Abstract:
  raw_layout: Layout
  lines: list[Line]
  did_update: bool = False

class Fragment:
  def __init__(self, page_index: int) -> None:
    self._page_index: int = page_index
    self._headlines: list[tuple[Layout, list[Line]]] = []
    self._abstracts: list[_Abstract] = []

  @property
  def page_index(self) -> int:
    return self._page_index

  @property
  def is_abstracts_empty(self) -> bool:
    return bool(self._abstracts)

  def append_headline(self, headline_layout: Layout) -> None:
    self._headlines.append((
      headline_layout,
      [
        Line(
          id=-1,
          text=line.text,
          splitted=False,
        )
        for line in headline_layout.lines
      ],
    ))

  def append_abstract_line(
        self,
        parent_layout: Layout,
        text: str,
        splitted: bool = False,
      ) -> None:

    to_append_abstract: _Abstract | None = None
    if self._abstracts:
      last_abstract = self._abstracts[-1]
      if last_abstract.raw_layout.id == parent_layout.id:
        to_append_abstract = last_abstract

    if to_append_abstract:
      to_append_abstract.lines.append(Line(
        id=-1,
        text=text,
        splitted=splitted,
      ))
    else:
      self._abstracts.append(_Abstract(
        raw_layout=parent_layout,
        did_update=False,
        lines=[Line(
          id=-1,
          text=text,
          splitted=splitted,
        )],
      ))

  def define_line_ids(self, first_id: int) -> int:
    next_id: int = first_id
    for line in self._lines():
      line.id = next_id
      next_id += 1
    return next_id

  def pop_line(self, line_id: int) -> Line | None:
    lines: list[Line] | None = None
    for _, lines in self._headlines:
      for i, line in enumerate(lines):
        if line.id == line_id:
          return lines.pop(i)

    for abstract in self._abstracts:
      for i, line in enumerate(abstract.lines):
        if line.id == line_id and not line.splitted:
          abstract.did_update = True
          return abstract.lines.pop(i)

    return None

  def update_headline(self, id: str, lines: list[Line]):
    for i, (headline, _) in enumerate(self._headlines):
      if headline.id == id:
        self._headlines[i] = (headline, lines)
        return

  def to_request_xml(self, id: int):
    fragment_element = Element("fragment")
    fragment_element.set("id", _to_abc_id(id))
    fragment_element.set("page-index", str(self._page_index))

    for headline, lines in self._headlines:
      headline_element = self._to_headline_xml(headline, lines)
      fragment_element.append(headline_element)

    if self._abstracts:
      abstract_element = Element("abstract")
      fragment_element.append(abstract_element)
      for abstract in self._abstracts:
        for line in abstract.lines:
          abstract_element.append(line.to_xml())

    return fragment_element

  def generate_patch_xmls(self) -> Generator[tuple[int, Element], None, None]:
    for headline, lines in self._headlines:
      headline_element = self._to_headline_xml(headline, lines)
      yield headline.page_index, headline_element

    for abstract in self._abstracts:
      if not abstract.did_update:
        continue
      raw_layout = abstract.raw_layout
      layout_element = Element(raw_layout.kind.value)
      layout_element.set("id", raw_layout.id)
      for line in abstract.lines:
        line_element = Element("line")
        line_element.text = line.text
        layout_element.append(line_element)
      yield raw_layout.page_index, layout_element

  def _lines(self) -> Generator[Line, None, None]:
    for _, lines in self._headlines:
      yield from lines
    for abstract in self._abstracts:
      yield from abstract.lines

  def _to_headline_xml(self, headline: Layout, lines: list[Line]) -> Element:
    headline_xml = Element(headline.kind.value)
    headline_xml.set("id", headline.id)
    for line in lines:
      headline_xml.append(line.to_xml())
    return headline_xml

class FragmentRequest:
  def __init__(self):
    self._fragments: list[Fragment] = []

  @property
  def fragments_count(self) -> int:
    return len(self._fragments)

  @property
  def begin_page_index(self) -> int:
    return min(f.page_index for f in self._fragments)

  @property
  def end_page_index(self) -> int:
    return max(f.page_index for f in self._fragments)

  def append(self, fragment: Fragment) -> None:
    self._fragments.append(fragment)

  def generate_matched_mapper(self, resp_xml: Element) -> Generator[tuple[str, int], None, None]:
    contents_map = resp_xml.find("contents-map")
    if contents_map is None:
      return
    for child in contents_map:
      if child.tag != "match":
        continue
      headline_id = child.get("headline-id", default=None)
      chapter_id = int(child.get("chapter-id", "-1"))
      if headline_id is None or chapter_id < 0:
        continue
      yield headline_id, chapter_id

  def complete_to_xml(self) -> Element:
    next_id: int = 1
    for fragment in self._fragments:
      next_id = fragment.define_line_ids(next_id)

    request_xml = Element("request")
    for id, fragment in enumerate(self._fragments):
      fragment_element = fragment.to_request_xml(id)
      request_xml.append(fragment_element)
    return request_xml

  def generate_patch_xmls(self, resp_xml: Element) -> Generator[tuple[int, Element], None, None]:
    patched_headline_indexes_set: set[int] = set()
    for index, headline_id, line_pairs in self._collect_headline_patch(resp_xml):
      if index >= len(self._fragments):
        continue
      fragment = self._fragments[index]
      lines: list[Line] = []
      for line_id, line_text in line_pairs:
        line = fragment.pop_line(line_id)
        if line is None:
          continue
        line.text = line_text
        lines.append(line)
      fragment.update_headline(headline_id, lines)
      patched_headline_indexes_set.add(index)

    for index in sorted(list(patched_headline_indexes_set)):
      fragment = self._fragments[index]
      yield from fragment.generate_patch_xmls()

  def _collect_headline_patch(self, resp_xml: Element):
    for child in resp_xml:
      if child.tag != "fixed-fragment":
        continue
      index = _parse_abc_id(child.get("id", "A"))
      if index >= len(self._fragments):
        continue
      for headline in child:
        headline_id = headline.get("id", None)
        if headline_id is None:
          continue
        lines: list[tuple[int, str]] = []
        for line in headline:
          if line.tag != "line":
            continue
          line_id = line.get("id", "-1")
          lines.append((int(line_id), line.text))
        yield index, headline_id, lines

def _to_abc_id(id: int) -> str:
  result = ""
  while id > 0:
    id, remainder = divmod(id, 26)
    result = chr(ord("A") + remainder) + result
  if not result:
    result = "A"
  return result

def _parse_abc_id(id: str) -> int:
  result = 0
  for char in id:
    result = result * 26 + (ord(char) - ord("A"))
  return result


from pathlib import Path
from xml.etree.ElementTree import Element

from ...llm import LLM
from ..reporter import Reporter, AnalysingStep
from ..utils import read_xml_file, Context
from .common import Phase, State
from .type import Contents, Chapter
from .collection import collect
from .utils import normalize_layout_xml


def extract_contents(
      llm: LLM,
      reporter: Reporter,
      workspace_path: Path,
      sequence_path: Path,
      max_data_tokens: int,
    ) -> Contents | None:

  context: Context[State] = Context(
    reporter=reporter,
    path=workspace_path,
    init=lambda: {
      "phase": Phase.INIT,
      "page_indexes": [],
      "max_data_tokens": max_data_tokens,
    },
  )
  if context.state["phase"] == Phase.NO_CONTENTS:
    return None

  extracted_path = context.path.joinpath("extracted.xml")
  extracted_xml: Element

  if context.state["phase"] == Phase.GENERATED:
    extracted_xml = read_xml_file(extracted_path)
  else:
    # why not do it in 1 step? Because in extreme cases the contents will be
    # very long (the contents will not be processed in sections), splitting it into
    # 2 steps can reduce the prompt length and help LLM focus.
    md_content = _extract_md_contents(
      llm=llm,
      context=context,
      sequence_path=sequence_path,
    )
    if md_content is None:
      context.state = {
        **context.state,
        "phase": Phase.NO_CONTENTS.value,
      }
      return None

    extracted_xml = llm.request_xml(
      template_name="contents/format",
      user_data=f"```Markdown\n{md_content}\n```",
    )
    context.write_xml_file(extracted_path, extracted_xml)
    context.state = {
      **context.state,
      "phase": Phase.GENERATED.value,
    }
    context.reporter.go_to_step(AnalysingStep.ANALYSE_CONTENTS)

  return _parse_contents_xml(context, extracted_xml)

def _extract_md_contents(llm: LLM, context: Context[State], sequence_path: Path) -> str | None:
  md_path = context.path.joinpath("extracted.md")
  md_content: str | None = None

  if md_path.exists():
    with md_path.open("r", encoding="utf-8") as f:
      md_content = f.read()
  else:
    context.reporter.go_to_step(AnalysingStep.COLLECT_CONTENTS)
    request_xml = Element("request")
    for paragraph in collect(llm, context, sequence_path):
      layout = normalize_layout_xml(paragraph)
      if layout is not None:
        request_xml.append(layout)
      context.reporter.increment()

    if len(request_xml) > 0:
      md_content = llm.request_markdown(
        template_name="contents/extractor",
        user_data=request_xml,
      )
      context.atomic_write(md_path, md_content)

  return md_content

def _parse_contents_xml(context: Context[State], contents_xml: Element) -> Contents:
  prefaces: list[Chapter] = []
  chapters: list[Chapter] = []

  for l0_child in contents_xml:
    l0_chapters: list[Chapter]
    if l0_child.tag == "prefaces":
      l0_chapters = prefaces
    elif l0_child.tag == "chapters":
      l0_chapters = chapters
    else:
      continue
    for chapter_xml in l0_child:
      chapter = _parse_chapter(chapter_xml)
      if chapter is not None:
        l0_chapters.append(chapter)

  next_id = 1
  contents = Contents(
    prefaces=prefaces,
    chapters=chapters,
    page_indexes=context.state["page_indexes"],
  )
  for chapter in contents:
    chapter.id = next_id
    next_id += 1

  return contents

def _parse_chapter(chapter_xml: Element) -> Chapter | None:
  if chapter_xml.tag != "chapter":
    return None

  name = (chapter_xml.text or "").strip()
  if not name:
    return None

  children: list[Chapter] = []
  for child in chapter_xml:
    sub_chapter = _parse_chapter(child)
    if sub_chapter is not None:
      children.append(sub_chapter)

  return Chapter(
    id=-1,
    name=name,
    children=children,
  )


from typing import Generator, Iterable
from xml.etree.ElementTree import Element

from .tag import Tag, TagKind
from .parser import parse_tags
from .transform import tag_to_element
from .utils import clone

# why implement XML decoding?
# https://github.com/oomol-lab/pdf-craft/issues/149
def decode_friendly(chars: Iterable[str], tags: Iterable[str] | str = ()) -> Generator[Element, None, None]:
  if isinstance(tags, str):
    tags = set((tags,))
  else:
    tags = set(tags)

  for element in _collect_elements(chars):
    if element.tag in tags or len(tags) == 0:
      yield clone(element)

def _collect_elements(chars: Iterable[str]) -> Generator[Element, None, None]:
  opening_stack: list[Element] = []
  last_closed_element: Element | None = None

  for cell in parse_tags(chars):
    if isinstance(cell, Tag):
      tag: Tag = cell
      element = tag_to_element(tag)
      if tag.kind == TagKind.CLOSING:
        popped = _pop_element(tag.name, opening_stack)
        if popped is not None:
          yield popped
          last_closed_element = popped
        elif last_closed_element is not None:
          _append_to_tail(last_closed_element, tag.proto)
      else:
        if opening_stack:
          opening_stack[-1].append(element)
        if tag.kind == TagKind.SELF_CLOSING:
          yield element
          last_closed_element = element
        elif tag.kind == TagKind.OPENING:
          opening_stack.append(element)
          last_closed_element = None

    elif last_closed_element is not None:
      _append_to_tail(last_closed_element, cell)

    elif opening_stack:
      opening_stack[-1].text = cell

def _append_to_tail(element: Element, text: str) -> None:
  if element.tail:
    element.tail += text
  else:
    element.tail = text

def _pop_element(tag_name: str, opening_stack: list[Element]) -> Element | None:
  index = -1
  for i in range(len(opening_stack) - 1, -1, -1):
    opening_element = opening_stack[i]
    if tag_name == opening_element.tag:
      index = i
      break
  if index == -1:
    return None

  popped: Element | None = None
  for _ in range(len(opening_stack) - index):
    popped = opening_stack.pop()
  return popped



from typing import Callable
from enum import auto, Enum
from dataclasses import dataclass
from PIL.Image import Image
from doc_page_extractor import Rectangle


# func(completed_pages: int, all_pages: int) -> None
PDFPageExtractorProgressReport = Callable[[int, int], None]

class OCRLevel(Enum):
  Once = auto()
  OncePerLayout = auto()

class ExtractedTableFormat(Enum):
  LATEX = auto()
  MARKDOWN = auto()
  HTML = auto()
  DISABLE = auto()

class TextKind(Enum):
  TITLE = 0
  PLAIN_TEXT = 1
  ABANDON = 2

@dataclass
class Text:
  content: str
  rank: float
  rect: Rectangle

@dataclass
class BasicBlock:
  rect: Rectangle
  texts: list[Text]
  font_size: float

@dataclass
class TextBlock(BasicBlock):
  kind: TextKind
  has_paragraph_indentation: bool = False
  last_line_touch_end: bool = False

class TableFormat(Enum):
  LATEX = auto()
  MARKDOWN = auto()
  HTML = auto()
  UNRECOGNIZABLE = auto()

@dataclass
class TableBlock(BasicBlock):
  content: str
  format: TableFormat
  image: Image

@dataclass
class FormulaBlock(BasicBlock):
  content: str | None
  image: Image

@dataclass
class FigureBlock(BasicBlock):
  image: Image

AssetBlock = TableBlock | FormulaBlock | FigureBlock
Block = TextBlock | AssetBlock


from __future__ import annotations
from typing import Generator
from xml.etree.ElementTree import Element

from ...xml import encode_friendly
from ...llm import LLM
from ..data import ASSET_LAYOUT_KINDS


class SequenceRequest:
  def __init__(self):
    self.begin: int = -1
    self.end: int = -1
    self.raw_pages: list[RawPage] = []

  def append(self, page_index: int, raw_page: RawPage):
    self.raw_pages.append(raw_page)
    self.end = page_index
    if self.begin == -1:
      self.begin = page_index

  def raw_page(self, page_index: int) -> RawPage | None:
    for raw_page in self.raw_pages:
      if raw_page.page_index == page_index:
        return raw_page
    return None

  def inject_ids_and_get_xml(self) -> Element:
    next_line_id: int = 1
    request_element = Element("request")

    for raw_page in self.raw_pages:
      page_element = Element("page")
      page_element.set("page-index", str(raw_page.page_index))
      request_element.append(page_element)
      next_asset_index: int = 0

      for layout_element in raw_page.children:
        page_element.append(layout_element)
        if layout_element.tag not in ASSET_LAYOUT_KINDS:
          for line_element in layout_element:
            if line_element.tag == "line":
              line_element.set("id", str(next_line_id))
              next_line_id += 1
        else:
          asset = raw_page.asset_datas[next_asset_index]
          asset_line = next((c for c in layout_element if c.tag == "line"), None)
          next_asset_index += 1
          if asset_line is not None:
            asset_line.set("id", str(next_line_id))
            asset.line_id = next_line_id
            next_line_id += 1

    return request_element

class RawPage:
  def __init__(self, raw_element: Element, page_index: int) -> None:
    self.asset_datas: list[_AssetData] = []
    self.children: list[Element] = []
    self.page_index: int = page_index

    for layout_element, asset_captions in self._handle_layout_elements(raw_element):
      if layout_element.tag in ASSET_LAYOUT_KINDS:
        asset_data = _AssetData(layout_element, asset_captions)
        self.asset_datas.append(asset_data)
        self.children.append(asset_data.element)
      else:
        self.children.append(layout_element)

    for layout_element in self.children:
      for line_element in layout_element:
        if line_element.tag == "line":
          # just as a placeholder, so that the tokens are as consistent as possible with the final output
          line_element.set("id", "X")

  def _handle_layout_elements(self, raw_element: Element) -> Generator[tuple[Element, list[Element]], None, None]:
    asset_and_captions: tuple[Element, list[Element]] | None = None

    for layout_element in raw_element:
      if layout_element.tag in ASSET_LAYOUT_KINDS:
        if asset_and_captions is not None:
          asset, captions = asset_and_captions
          yield asset, captions
        asset_and_captions = (layout_element, [])

      else:
        did_append_as_caption = False
        if asset_and_captions is not None:
          asset, captions = asset_and_captions
          if layout_element.tag == f"{asset.tag}-caption":
            captions.append(layout_element)
            did_append_as_caption = True

        if not did_append_as_caption:
          if asset_and_captions is not None:
            asset, captions = asset_and_captions
            asset_and_captions = None
            yield asset, captions
          yield layout_element, []
          asset_and_captions = None

    if asset_and_captions is not None:
      asset, captions = asset_and_captions
      yield asset, captions

  def tokens_count(self, llm: LLM) -> int:
    tokens_count: int = 0
    for element in self.children:
      text = encode_friendly(element)
      tokens_count += len(llm.encode_tokens(text))
    return tokens_count

  def inject_assets(self, line_ids: list[int]) -> Generator[tuple[int, Element | None], None, None]:
    if len(line_ids) <= 1:
      for line_id in line_ids:
        yield line_id, None
      return

    pre_line_id: int | None = None
    asset_datas = [
      asset for _, asset in sorted(
        [
          (asset.line_id, asset)
          for asset in self.asset_datas
          if asset.line_id is not None
        ],
        key=lambda x: x[0],
      )
    ]
    for line_id in sorted(line_ids):
      did_yield_this = False
      for asset in asset_datas:
        if asset.line_id == line_id:
          yield line_id, asset.to_saved_xml()
          did_yield_this = True
          break
        if pre_line_id is not None and \
           pre_line_id < asset.line_id < line_id:
          yield asset.line_id, asset.to_saved_xml()
      if not did_yield_this:
        yield line_id, None
      pre_line_id = line_id

  def assets_in_range(
        self,
        after_line_id: int | None = None,
        before_line_id: int | None = None,
      ) -> Generator[Element, None, None]:
    for asset in self.asset_datas:
      line_id = asset.line_id
      if line_id is None:
        continue
      if after_line_id is not None and line_id <= after_line_id:
        continue
      if before_line_id is not None and line_id >= before_line_id:
        break
      yield asset.to_saved_xml()

class _AssetData:
  def __init__(self, element: Element, captions: list[Element]):
    self.element = element
    self.line_id: int | None = None
    self._captions: list[Element] = captions
    self._hash: str | None = element.attrib.pop("hash", None)

    if element.tag == "figure":
      element.clear()
      element.append(self._create_line("[[OCR recognized figure here]]"))

    elif element.tag == "table":
      element.clear()
      element.append(self._create_line("[[OCR recognized table here]]"))

    elif element.tag == "formula":
      latex_element = element.find("latex")
      line_text = "[[OCR recognized formula here]]"
      if latex_element is not None and latex_element.text:
        line_text = latex_element.text

      element.clear()
      element.append(self._create_line(line_text))

  def to_saved_xml(self) -> Element:
    asset_element = Element(self.element.tag)
    if asset_element.tag == "formula":
      line_element = self.element.find("line")
      if line_element is not None:
        latex_element = Element("latex")
        latex_element.text = line_element.text
        asset_element.append(latex_element)

    for raw_caption_element in self._captions:
      caption_element = Element("caption")
      caption_element.extend(raw_caption_element)
      asset_element.append(caption_element)

    if self._hash is not None:
      asset_element.set("hash", self._hash)
    return asset_element

  def _create_line(self, text: str) -> Element:
    line_element = Element("line")
    line_element.text = text
    line_element.attrib["confidence"] = "1.0"
    return line_element


from pathlib import Path
from typing import Generator

from ..data import Paragraph, Layout, Caption, Line
from ..sequence import read_paragraphs
from .mark import transform2mark, Mark


ExtractedFootnote = tuple[int, Mark, list[Layout]]

def extract_footnote_references(footnote_path: Path) -> Generator[ExtractedFootnote, None, None]:
  pre_page_index: int = -1
  last: ExtractedFootnote | None = None

  for page_index, mark, layouts in _extract_mark_and_layouts(footnote_path):
    if mark is not None:
      if last is not None:
        yield last
      last = (page_index, mark, layouts)
    elif last is not None:
      last_page_index, _, last_layouts = last
      if page_index == last_page_index:
        last_layouts.extend(layouts)
      elif pre_page_index == page_index - 1:
        # this indicates that the footnote of the previous page will continue after turning the page
        last_layouts.extend(layouts)
      else:
        yield last
        last = None
        # TODO: 此时当前 layouts 就会被抛弃。此处代码逻辑上无解，可以考虑交给 LLM 来处理
    pre_page_index = page_index

  if last is not None:
    yield last

def _extract_mark_and_layouts(footnote_path: Path) -> Generator[tuple[int, Mark | None, list[Layout]], None, None]:
  for paragraph in read_paragraphs(footnote_path):
    for mark, lines in _extract_footnote_from_paragraph(paragraph):
      layouts = _transform2layouts(lines)
      page_index = layouts[0].page_index
      yield page_index, mark, layouts

def _extract_footnote_from_paragraph(paragraph: Paragraph) -> Generator[tuple[Mark | None, list[tuple[Layout, Line]]], None, None]:
  current_mark: Mark | None = None
  lines: list[tuple[Layout, Line]] = []

  for layout in paragraph.layouts:
    for line in layout.lines:
      text = line.text.strip()
      mark = text and transform2mark(text[0])
      if isinstance(mark, Mark):
        if lines:
          yield current_mark, lines
          lines = []
        current_mark = mark
        text = text[1:].strip()
      lines.append((layout, Line(
        text=text,
        confidence=line.confidence,
      )))

  if lines:
    yield current_mark, lines

def _transform2layouts(lines: list[tuple[Layout, Line]]) -> list[Layout]:
  layouts: dict[str, tuple[int, Layout]] = {}
  for layout, line in lines:
    lines: list[Line]
    if layout.id not in layouts:
      lines = []
      layouts[layout.id] = (
        len(layouts),
        Layout(
          kind=layout.kind,
          page_index=layout.page_index,
          order_index=layout.order_index,
          caption=Caption(lines=[]),
          lines=lines,
        ),
      )
    else:
      lines = layouts[layout.id][1].lines
    lines.append(line)

  return list(
    line for _, line in sorted(
      list(layouts.values()),
      key=lambda x: x[0],
    )
  )


import shutil

from pathlib import Path
from xml.etree.ElementTree import Element

from ..data import Paragraph, AssetLayout
from ..utils import Partition, PartitionTask
from .common import State, Corrector
from .repeater import repeat_multiple_correct
from .paragraphs_reader import ParagraphsReader


class MultipleCorrector(Corrector):

  def do(self, from_path: Path, request_path: Path, is_footnote: bool) -> None:
    request_path.mkdir(parents=True, exist_ok=True)
    self.ctx.reporter.set(
      max_count=sum(1 for _ in self.generate_request_xml(from_path)),
    )
    partition: Partition[tuple[int, int], State, Element] = Partition(
      dimension=2,
      context=self.ctx,
      sequence=self.generate_request_xml(from_path),
      done=lambda _, __: self.ctx.reporter.increment(),
      remove=lambda begin, end: shutil.rmtree(
        request_path / _file_name("steps", begin, end),
      ),
    )
    with partition:
      self.threads.run(
        next_task=partition.pop_task,
        thread_payload=lambda: ParagraphsReader(from_path),
        invoke=lambda reader, task: self._emit_request(
          task=task,
          reader=reader,
          request_path=request_path,
          is_footnote=is_footnote,
        ),
      )

  def _emit_request(
        self,
        task: PartitionTask[tuple[int, int], State, Element],
        reader: ParagraphsReader,
        request_path: Path,
        is_footnote: bool,
      ) -> Element:

    with task:
      begin = task.begin
      end = task.end
      request_element = task.payload
      resp_element = repeat_multiple_correct(
        llm=self.llm,
        context=self.ctx,
        save_path=request_path / _file_name("steps",begin, end),
        raw_request=request_element,
        is_footnote=is_footnote,
      )
      self._apply_updation(
        reader=reader,
        request_path=request_path,
        request_element=request_element,
        resp_element=resp_element,
      )

  def _apply_updation(
        self,
        request_path: Path,
        reader: ParagraphsReader,
        request_element: Element,
        resp_element: Element,
      ) -> None:

    raw_lines_list = list(self.extract_lines(request_element))
    if not raw_lines_list:
      return

    corrected_lines_dict = dict(self.extract_lines(resp_element))
    paragraphs: dict[tuple[int, int], Paragraph] = {}

    for index, raw_lines in sorted(raw_lines_list, key=lambda x: x[0]):
      page_index, order_index = index
      paragraph = reader.read(
        layout_index=(page_index, order_index),
      )
      if paragraph is None:
        continue

      para_index = (paragraph.page_index, paragraph.order_index)
      if para_index not in paragraphs:
        paragraphs[para_index] = paragraph

      corrected_lines = corrected_lines_dict.get(index, None)
      if corrected_lines is None:
        corrected_lines = raw_lines

      self._apply_paragraph_lines(
        paragraph=paragraph,
        lines=corrected_lines,
      )

    chunk_element = Element("chunk")

    for para_index in sorted(list(paragraphs.keys())):
      page_index, order_index = para_index
      paragraph = paragraphs[para_index]
      if not paragraph.layouts:
        continue
      paragraph_element = paragraph.to_xml()
      paragraph_element.set("page-index", str(page_index))
      paragraph_element.set("order-index", str(order_index))
      chunk_element.append(paragraph_element)

    begin, _ = raw_lines_list[0]
    end, _ = raw_lines_list[-1]
    file_name = _file_name("chunk", begin, end) + ".xml"
    self.ctx.write_xml_file(
      file_path=request_path / file_name,
      xml=chunk_element,
    )

  def _apply_paragraph_lines(self, paragraph: Paragraph, lines: list[str]):
    next_line_index: int = 0
    limit_length: int = -1

    for i, layout in enumerate(paragraph.layouts):
      for j, line in enumerate(layout.lines):
        if next_line_index >= len(lines):
          limit_length = j
          break
        line.text = lines[next_line_index]
        next_line_index += 1

      if limit_length < 0:
        continue
      if isinstance(layout, AssetLayout):
        continue

      layout.lines = layout.lines[:limit_length]
      if layout.lines:
        limit_length = i + 1
      else:
        limit_length = i

    if limit_length != -1:
      paragraph.layouts = paragraph.layouts[:limit_length]

def _file_name(name: str, begin: tuple[int, int], end: tuple[int, int]) -> str:
  return f"{name}_{begin[0]}_{begin[1]}_{end[0]}_{end[1]}"


import re

from pathlib import Path
from ...llm import LLM
from ..sequence import decode_paragraph, ParagraphWriter
from ..reporter import Reporter, AnalysingStep
from ..utils import read_xml_file, Context, MultiThreads
from .common import State, Phase, Level, Corrector
from .single_corrector import SingleCorrector
from .multiple_corrector import MultipleCorrector


def correct(
      llm: LLM,
      reporter: Reporter,
      threads: MultiThreads,
      level: Level,
      workspace_path: Path,
      text_path: Path,
      footnote_path: Path,
      max_data_tokens: int,
    ) -> Path:

  context: Context[State] = Context(
    reporter=reporter,
    path=workspace_path,
    init=lambda: {
      "phase": Phase.Text.value,
      "level": level.value,
      "max_data_tokens": max_data_tokens,
      "completed_ranges": [],
    },
  )
  corrector: Corrector
  output_path = workspace_path / "output"
  text_request_path = workspace_path / "text"
  footnote_request_path = workspace_path / "footnote"

  if context.state["level"] == Level.Single:
    corrector = SingleCorrector(llm, context, threads)
  elif context.state["level"] == Level.Multiple:
    corrector = MultipleCorrector(llm, context, threads)
  else:
    raise ValueError(f"Unknown level: {context.state['level']}")

  if context.state["phase"] == Phase.Text:
    if text_path.exists():
      reporter.go_to_step(AnalysingStep.CORRECT_TEXT)
      corrector.do(
        from_path=text_path,
        request_path=text_request_path,
        is_footnote=False,
      )
    context.state = {
      **context.state,
      "phase": Phase.FOOTNOTE.value,
      "completed_ranges": [],
    }

  if context.state["phase"] == Phase.FOOTNOTE:
    if footnote_path.exists():
      reporter.go_to_step(AnalysingStep.CORRECT_FOOTNOTE)
      corrector.do(
        from_path=footnote_path,
        request_path=footnote_request_path,
        is_footnote=True,
      )
    context.state = {
      **context.state,
      "phase": Phase.GENERATION.value,
      "completed_ranges": [],
    }

  if context.state["phase"] == Phase.GENERATION:
    if text_request_path.exists():
      _generate_paragraph_files(
        context=context,
        request_path=text_request_path,
        output_path=output_path / "text",
      )
    if footnote_request_path.exists():
      _generate_paragraph_files(
        context=context,
        request_path=footnote_request_path,
        output_path=output_path / "footnote",
      )
    context.state = {
      **context.state,
      "phase": Phase.COMPLETED.value,
    }
  return output_path

_CHUNK_FILE_PATTERN = re.compile(r"^chunk(_\d+){4}\.xml$")
_CHUNK_FILE_HEAD_AND_TAIL_PATTERN = re.compile(r"(^chunk_|\.xml$)")

def _generate_paragraph_files(context: Context[State], request_path: Path, output_path: Path):
  index_and_file_list: list[tuple[tuple[int, int], tuple[int, int], Path]] = []
  for file in request_path.iterdir():
    matches = re.match(_CHUNK_FILE_PATTERN, file.name)
    if not matches:
      continue
    indexes_text = re.sub(_CHUNK_FILE_HEAD_AND_TAIL_PATTERN, "", file.name)
    indexes = list(int(i) for i in indexes_text.split("_"))
    index1 = (indexes[0], indexes[1])
    index2 = (indexes[2], indexes[3])
    index_and_file_list.append((index1, index2, file))

  writer = ParagraphWriter(context, output_path)
  for _, _, file in sorted(index_and_file_list, key=lambda x: x[0]):
    for paragraph_element in read_xml_file(file):
      paragraph = decode_paragraph(
        element=paragraph_element,
        page_index=int(paragraph_element.get("page-index", "-1")),
        order_index=int(paragraph_element.get("order-index", "-1")),
      )
      writer.write(paragraph)



from dataclasses import dataclass
from typing import Generator
from pathlib import Path
from xml.etree.ElementTree import Element

from ...llm import LLM
from ...xml import encode
from ..utils import Context
from ..sequence import read_paragraphs
from ..data import Paragraph, ParagraphType
from .common import Phase, State
from .utils import normalize_layout_xml


def collect(llm: LLM, context: Context[State], sequence_path: Path) -> Generator[Paragraph, None, None]:
  yield from _Collector(llm, context, sequence_path).do()

@dataclass
class _Page:
  page_type: ParagraphType
  page_index: int
  paragraphs: list[Paragraph]

  def xml(self) -> Element | None:
    page = Element("page")
    page.set("page-index", str(self.page_index))

    for paragraph in self.paragraphs:
      merged_layout = normalize_layout_xml(paragraph)
      if merged_layout is not None:
        page.append(merged_layout)

    return page

class _Collector:
  def __init__(self, llm: LLM, context: Context[State], sequence_path: Path):
    self._llm: LLM = llm
    self._context: Context[State] = context
    self._sequence_path: Path = sequence_path

  def do(self) -> Generator[Paragraph, None, None]:
    phase = Phase(self._context.state["phase"])
    page_indexes = self._context.state["page_indexes"]

    if phase == Phase.INIT:
      page_indexes = []
      for page in self._read_pages():
        if page.page_type == ParagraphType.CONTENTS:
          page_indexes.append(page.page_index)
        else:
          break
      phase = Phase.COLLECT
      self._context.state = {
        **self._context.state,
        "phase": phase.value,
        "page_indexes": page_indexes,
      }
    if not page_indexes:
      return

    if phase == Phase.COLLECT:
      for page in self._read_and_identify_pages(page_indexes):
        if page.page_index not in page_indexes:
          page_indexes.append(page.page_index)
          self._context.state = {
            **self._context.state,
            "page_indexes": page_indexes,
          }
        yield from page.paragraphs

      phase = Phase.ANALYSE
      self._context.state = {
        **self._context.state,
        "phase": phase.value,
      }
    else:
      for page, matched in self._read_pages_with_matched(page_indexes):
        if not matched:
          break
        yield from page.paragraphs

  def _read_and_identify_pages(self, page_indexes: list[int]) -> Generator[_Page, None, None]:
    yield_pages_count: int = 0
    unmatched_pages: list[_Page] = []
    unmatched_elements: list[Element] = []
    unmatched_tokens: int = 0

    def identify_matched_pages_and_check_done() -> tuple[list[_Page], bool]:
      nonlocal yield_pages_count, unmatched_pages, unmatched_elements, unmatched_tokens
      if not unmatched_elements:
        return [], False

      matched_pages: list[_Page] = []
      matched_result = self._check_pages_matched(
        yield_pages_count=yield_pages_count,
        page_elements=unmatched_elements,
      )
      for page, matched in zip(unmatched_pages, matched_result):
        if not matched:
          return matched_pages, True
        matched_pages.append(page)

      unmatched_pages.clear()
      unmatched_elements.clear()
      unmatched_tokens = 0
      return matched_pages, False

    max_data_tokens = self._context.state["max_data_tokens"]

    for page, matched in self._read_pages_with_matched(page_indexes):
      if matched or page.page_type == ParagraphType.CONTENTS:
        if unmatched_pages:
          matched_pages, done = identify_matched_pages_and_check_done()
          yield from matched_pages
          if done:
            return
          yield_pages_count += len(matched_pages)
        yield page
        yield_pages_count += 1
      else:
        page_element = page.xml()
        page_tokens = len(self._llm.encode_tokens(encode(page_element)))
        if unmatched_pages and unmatched_tokens + page_tokens > max_data_tokens:
          matched_pages, done = identify_matched_pages_and_check_done()
          yield from matched_pages
          if done:
            return
          yield_pages_count += len(matched_pages)

        unmatched_pages.append(page)
        unmatched_elements.append(page_element)
        unmatched_tokens += page_tokens

    if unmatched_pages:
      matched_pages, _ = identify_matched_pages_and_check_done()
      yield from matched_pages

  def _read_pages_with_matched(self, page_indexes: list[int]) -> Generator[tuple[_Page, bool], None, None]:
    page_indexes.sort()
    max_page_index = -1
    if page_indexes:
      max_page_index = max(page_indexes)
    for page in self._read_pages():
      if page.page_index in page_indexes:
        yield page, True
      elif page.page_index > max_page_index:
        yield page, False

  def _read_pages(self) -> Generator[_Page, None, None]:
    page: _Page | None = None
    skip_not_matched = True

    for paragraph in read_paragraphs(self._sequence_path):
      if len(paragraph.layouts) == 0:
        continue
      elif paragraph.type == ParagraphType.CONTENTS:
        skip_not_matched = False
      elif skip_not_matched:
        continue

      if page is not None:
        if page.page_index == paragraph.page_index:
          page.paragraphs.append(paragraph)
        else:
          yield page
          page = None

      if page is None:
        page = _Page(
          page_type=paragraph.type,
          page_index=paragraph.page_index,
          paragraphs=[paragraph],
        )
    if page is not None:
      yield page

  def _check_pages_matched(self, yield_pages_count: int, page_elements: list[Element]) -> list[bool]:
    request_xml = Element("request")
    request_xml.extend(page_elements)
    resp_xml = self._llm.request_xml(
      template_name="contents/identifier",
      user_data=request_xml,
      params={
        "last_pages_count": yield_pages_count,
        "current_pages_count": len(page_elements),
      },
    )
    matched_page_indexes: set[int] = set()
    matched: list[bool] = []

    for resp_page in resp_xml:
      page_index = resp_page.get("page-index", None)
      page_type = resp_page.get("type", None)
      if page_index is not None and ParagraphType.CONTENTS == page_type:
        matched_page_indexes.add(page_index)

    for raw_page in page_elements:
      page_index = raw_page.get("page-index", None)
      matched.append(page_index in matched_page_indexes)

    return matched



import re
import io

from enum import Enum
from unicodedata import category
from alphabet_detector import AlphabetDetector


def check_texts_matching_rate(text1: str, text2: str) -> tuple[float, int]:
  words1: list[str] = list(split_into_words(text1))
  words2: list[str] = list(split_into_words(text2))

  if len(words1) > len(words2):
    words1, words2 = words2, words1

  marked_taken_list: list[bool] = [False] * len(words2)
  matched_indexes: list[int] = []
  for word1 in words1:
    matched_index: int = -1
    for i, word2 in enumerate(words2):
      if not marked_taken_list[i] and word1 == word2:
        marked_taken_list[i] = True
        matched_index = i
        break
    matched_indexes.append(matched_index)

  not_matched_count: int = len(words2) - len(words1)
  for i, matched_index in enumerate(matched_indexes):
    if matched_index == -1 or i > matched_index:
      not_matched_count += 1

  return 1.0 - not_matched_count / len(words2), len(words2)

class _Phase(Enum):
  Init = 0,
  Letter = 1,
  Character = 2,
  Number = 3,
  Space = 4,

def split_into_words(text: str):
  space_pattern = re.compile(r"\s")
  number_pattern = re.compile(r"\d")
  number_signs_pattern = re.compile(r"[\.,']")
  word_buffer = io.StringIO()
  phase: _Phase = _Phase.Init

  for char in text:
    if _is_letter(char):
      if phase == _Phase.Number:
        yield word_buffer.getvalue()
        word_buffer = io.StringIO()
      word_buffer.write(char)
      phase = _Phase.Letter

    elif number_pattern.match(char):
      if phase == _Phase.Letter:
        yield word_buffer.getvalue()
        word_buffer = io.StringIO()
      word_buffer.write(char)
      phase = _Phase.Letter

    elif phase == _Phase.Number and \
         number_signs_pattern.match(char):
      word_buffer.write(char)

    else:
      if phase == _Phase.Letter or \
        phase == _Phase.Number:
        yield word_buffer.getvalue()
        word_buffer = io.StringIO()

      if space_pattern.match(char):
        phase = _Phase.Space
      else: # others (like Chinese character)
        yield char
        phase = _Phase.Character

  if phase == _Phase.Letter:
    yield word_buffer.getvalue()

ad = AlphabetDetector()

# check is Latin, Cyrillic, Greek, or Hebrew letter
def _is_letter(char: str):
  if not category(char).startswith("L"):
    return False

  # AlphabetDetector unable to process punctuation
  return ad.is_latin(char) or \
         ad.is_cyrillic(char) or \
         ad.is_greek(char) or \
         ad.is_hebrew(char)


from typing import Generator
from dataclasses import dataclass
from enum import IntEnum
from xml.etree.ElementTree import Element
from resource_segmentation import split, Resource, Segment, Group
from ..data import ParagraphType


class TruncationKind(IntEnum):
  NO = -1
  VERIFIED = 1
  UNCERTAIN = 0

@dataclass
class _Child:
  page_index: int
  element: Element
  tokens: int
  tail: TruncationKind

class ParagraphDraft:
  def __init__(self, type: ParagraphType):
    self._type: ParagraphType = type
    self._children: list[_Child] = []
    self._tokens: int = 0

  @property
  def page_index(self) -> int:
    return self._children[0].page_index

  @property
  def type(self) -> ParagraphType:
    return self._type

  @property
  def tokens(self) -> int:
    return self._tokens

  def append(self, page_index: int, element: Element, tokens: int):
    self._tokens += tokens
    self._children.append(_Child(
      page_index=page_index,
      element=element,
      tokens=tokens,
      tail=TruncationKind.NO,
    ))

  def set_tail_truncation(self, kind: TruncationKind):
    self._children[-1].tail = kind

  def fork(self, max_chunk_tokens: int) -> Generator["ParagraphDraft", None, None]:
    pre_incision: TruncationKind = TruncationKind.NO
    resources: list[Resource[_Child]] = []

    for child in self._children:
      incision = child.tail
      resources.append(Resource(
        count=child.tokens,
        start_incision=pre_incision,
        end_incision=incision,
        payload=child,
      ))
      pre_incision = incision

    for group in split(
      resources=iter(resources),
      max_segment_count=max_chunk_tokens,
      border_incision=TruncationKind.NO,
      gap_rate=0,
      tail_rate=0,
    ):
      children = [r.payload for r in self._iter_group_body(group)]
      if children:
        forked = ParagraphDraft(self._type)
        for child in children:
          forked.append(
            page_index=child.page_index,
            element=child.element,
            tokens=child.tokens,
          )
        forked.set_tail_truncation(children[-1].tail)
        yield forked

  def to_xml(self) -> Element:
    element = Element("paragraph")
    element.set("type", self._type.value)
    for child in self._children:
      element.append(child.element)
    return element

  def _iter_group_body(self, group: Group[_Child]):
    for item in group.body:
      if isinstance(item, Resource):
        yield item
      elif isinstance(item, Segment):
        for resource in item.resources:
          yield resource